import pygame
from pygame import gfxdraw

from pygame_widgets.widget import WidgetBase


class Toggle(WidgetBase):
    def __init__(self, win, x, y, width, height, **kwargs):
        super().__init__(win, x, y, width, height)

        self.value = kwargs.get('startOn', False)
        self.onColour = kwargs.get('onColour', (141, 185, 244))
        self.offColour = kwargs.get('offColour', (150, 150, 150))
        self.handleOnColour = kwargs.get('handleOnColour', (26, 115, 232))
        self.handleOffColour = kwargs.get('handleOffColour', (200, 200, 200))

        self.handleRadius = kwargs.get('handleRadius', int(self._height / 1.3))
        self.radius = self._height // 2

        self.hidden = False
        self.clicked = False
        self.colour = self.onColour if self.value else self.offColour
        self.handleColour = self.handleOnColour if self.value else self.handleOffColour

    def toggle(self):
        self.value = not self.value
        self.colour = self.onColour if self.value else self.offColour
        self.handleColour = self.handleOnColour if self.value else self.handleOffColour

    def listen(self, events):
        if not self._hidden:
            pressed = pygame.mouse.get_pressed()[0]
            x, y = pygame.mouse.get_pos()

            if self.contains(x, y):
                if pressed:
                    if not self.clicked:
                        self.clicked = True
                        self.toggle()

                elif self.clicked:
                    self.clicked = False

            elif not pressed:
                self.clicked = False

    def draw(self):
        pygame.draw.rect(self.win, self.colour, (self._x, self._y, self._width, self._height))

        pygame.draw.circle(self.win, self.colour, (self._x, self._y + self._height // 2), self.radius)
        pygame.draw.circle(self.win, self.colour, (self._x + self._width, self._y + self._height // 2), self.radius)

        circle = (
            self._x + (
                self._width - self.handleRadius + self.radius if self.value else self.handleRadius - self.radius
            ),
            self._y + self._height // 2
        )

        gfxdraw.filled_circle(self.win, *circle, self.handleRadius, self.handleColour)
        gfxdraw.aacircle(self.win, *circle, self.handleRadius, self.handleColour)

    def getValue(self):
        return self.value


if __name__ == '__main__':
    pygame.init()
    win = pygame.display.set_mode((1000, 600))

    toggle = Toggle(win, 100, 100, 100, 40)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        toggle.listen(events)
        toggle.draw()

        pygame.display.update()



from typing import Callable
import pygame

from pygame_widgets.widget import WidgetBase


class ProgessBar(WidgetBase):
    def __init__(self, win, x, y, width, height, progress: Callable[[], float], **kwargs):
        super().__init__(win, x, y, width, height)
        self.progress = progress

        self.curved = kwargs.get('curved', False)

        self.completedColour = kwargs.get('completedColour', (0, 200, 0))
        self.incompletedColour = kwargs.get('incompletedColour', (100, 100, 100))

        self.percent = self.progress()

        self.radius = self._height / 2 if self.curved else 0

    def listen(self, events):
        pass

    def draw(self):
        """ Display to surface """
        self.percent = min(max(self.progress(), 0), 1)

        if not self._hidden:
            if self.curved:
                if self.percent == 0:
                    pygame.draw.circle(self.win, self.incompletedColour,
                                       (self._x, self._y + self._height // 2), self.radius)
                    pygame.draw.circle(self.win, self.incompletedColour,
                                       (self._x + self._width, self._y + self._height // 2),
                                       self.radius)
                elif self.percent == 1:
                    pygame.draw.circle(self.win, self.completedColour,
                                       (self._x, self._y + self._height // 2), self.radius)
                    pygame.draw.circle(self.win, self.completedColour,
                                       (self._x + self._width, self._y + self._height // 2),
                                       self.radius)
                else:
                    pygame.draw.circle(self.win, self.completedColour, (self._x, self._y + self._height // 2),
                                       self.radius)
                    pygame.draw.circle(self.win, self.incompletedColour,
                                       (self._x + self._width, self._y + self._height // 2),
                                       self.radius)

            pygame.draw.rect(self.win, self.completedColour,
                             (self._x, self._y, int(self._width * self.percent), self._height))
            pygame.draw.rect(self.win, self.incompletedColour,
                             (self._x + int(self._width * self.percent), self._y,
                              int(self._width * (1 - self.percent)), self._height))


if __name__ == '__main__':
    import time

    startTime = time.time()

    pygame.init()
    win = pygame.display.set_mode((1000, 600))

    progressBar = ProgessBar(win, 100, 100, 500, 40, lambda: 1 - (time.time() - startTime) / 10, curved=True)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        progressBar.draw()

        pygame.display.update()



import pygame
from pygame import gfxdraw
import math

from pygame_widgets.widget import WidgetBase
from pygame_widgets import TextBox


class Slider(WidgetBase):
    def __init__(self, win, x, y, width, height, **kwargs):
        super().__init__(win, x, y, width, height)

        self.selected = False

        self.min = kwargs.get('min', 0)
        self.max = kwargs.get('max', 99)
        self.step = kwargs.get('step', 1)

        self.colour = kwargs.get('colour', (200, 200, 200))
        self.handleColour = kwargs.get('handleColour', (0, 0, 0))

        self.borderThickness = kwargs.get('borderThickness', 3)
        self.borderColour = kwargs.get('borderColour', (0, 0, 0))

        self.value = self.round(kwargs.get('initial', (self.max + self.min) / 2))
        self.value = max(min(self.value, self.max), self.min)

        self.curved = kwargs.get('curved', True)

        self.vertical = kwargs.get('vertical', False)

        if self.curved:
            if self.vertical:
                self.radius = self._width // 2
            else:
                self.radius = self._height // 2

        if self.vertical:
            self.handleRadius = kwargs.get('handleRadius', int(self._width / 1.3))
        else:
            self.handleRadius = kwargs.get('handleRadius', int(self._height / 1.3))

    def listen(self, events):
        pressed = pygame.mouse.get_pressed()[0]
        x, y = pygame.mouse.get_pos()

        if pressed:
            if self.contains(x, y):
                self.selected = True

            if self.selected:
                if self.vertical:
                    self.value = self.max - self.round((y - self._y) / self._height * self.max)
                    self.value = max(min(self.value, self.max), self.min)
                else:
                    self.value = self.round((x - self._x) / self._width * self.max + self.min)
                    self.value = max(min(self.value, self.max), self.min)

        else:
            self.selected = False

    def draw(self):
        pygame.draw.rect(self.win, self.colour, (self._x, self._y, self._width, self._height))

        if self.vertical:
            if self.curved:
                pygame.draw.circle(self.win, self.colour, (self._x + self._width // 2, self._y), self.radius)
                pygame.draw.circle(self.win, self.colour, (self._x + self._width // 2, self._y + self._height), self.radius)
            circle = (self._x + self._width // 2,
                      int(self._y + (self.max - self.value) / (self.max - self.min) * self._height))
        else:
            if self.curved:
                pygame.draw.circle(self.win, self.colour, (self._x, self._y + self._height // 2), self.radius)
                pygame.draw.circle(self.win, self.colour, (self._x + self._width, self._y + self._height // 2), self.radius)
            circle = (int(self._x + (self.value - self.min) / (self.max - self.min) * self._width),
                      self._y + self._height // 2)

        gfxdraw.filled_circle(self.win, *circle, self.handleRadius, self.handleColour)
        gfxdraw.aacircle(self.win, *circle, self.handleRadius, self.handleColour)

    def contains(self, x, y):
        if self.vertical:
            handleX = self._x + self._width // 2
            handleY = int(self._y + (self.max - self.value) / (self.max - self.min) * self._height)
        else:
            handleX = int(self._x + (self.value - self.min) / (self.max - self.min) * self._width)
            handleY = self._y + self._height // 2

        if math.sqrt((handleX - x) ** 2 + (handleY - y) ** 2) <= self.handleRadius:
            return True

        return False

    def round(self, value):
        return self.step * round(value / self.step)

    def getValue(self):
        return self.value

    def setValue(self, value):
        self.value = value


if __name__ == '__main__':
    pygame.init()
    win = pygame.display.set_mode((1000, 600))

    slider = Slider(win, 100, 100, 800, 40, min=0, max=99, step=1)
    output = TextBox(win, 475, 200, 50, 50, fontSize=30)

    v_slider = Slider(win, 900, 200, 40, 300, min=0, max=99, step=1, vertical=True)
    v_output = TextBox(win, 800, 320, 50, 50, fontSize=30)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        slider.listen(events)
        v_slider.listen(events)

        slider.draw()
        v_slider.draw()

        output.setText(slider.getValue())
        v_output.setText(v_slider.getValue())

        output.draw()
        v_output.draw()

        pygame.display.update()



from pygame_widgets.textbox import TextBox
from pygame_widgets.button import Button, ButtonArray
from pygame_widgets.slider import Slider
from pygame_widgets.animations.animation import AnimationBase, Translate, Resize



class InvalidParameter(Exception):
    """ Exception called if an invalid parameter is passed into an animation"""


class InvalidParameterType(Exception):
    """ Exception called if an invalid parameter type is passed into an animation"""



from threading import Thread
import time
import pygame

from pygame_widgets.exceptions.exceptions import InvalidParameter, InvalidParameterType


class AnimationBase:
    def __init__(self, widget, timeout, allowMultiple=False, **kwargs):
        """Base for animations

        :param widget: The widget that the animation targets
        :param timeout: The time of the animation in seconds
        :param kwargs: The target of the animation, e.g. x=10 changes x position to 10
        """
        self.widget = widget
        self.timeout = timeout
        self.allowMultiple = allowMultiple
        self.params = kwargs
        self.thread = Thread()

        self.started = False
        self.runOnce = False

        self.checkValidParams()

    def checkValidParams(self):
        for param, target in self.params.items():
            value = self.widget.get(param)
            if value is None:
                raise InvalidParameter(
                    f'The parameter <{param}> is not a valid attribute of type {type(self.widget)}'
                )
            elif type(value) != type(target):
                raise InvalidParameterType(
                    f'Expected parameter <{param}> to be of type {type(value)} but found type {type(target)}'
                )

    def start(self):
        if not self.started and not (self.runOnce and not self.allowMultiple):
            self.thread = Thread(target=self.loop)
            self.thread.start()

    def loop(self):
        self.started = self.runOnce = True

        start = time.time()

        initialNumberParams = {}
        initialTupleParams = {}
        for param, target in self.params.items():
            initialValue = self.widget.get(param)
            if isinstance(initialValue, (int, float)):
                initialNumberParams[param] = initialValue
            elif isinstance(initialValue, (tuple, list)):
                initialTupleParams[param] = tuple(initialValue)

        # Animate
        while time.time() - start < self.timeout:
            step = (time.time() - start) / self.timeout

            # Numeric animation
            for param, initialValue in initialNumberParams.items():
                target = self.params[param]
                newValue = initialValue + step * (target - initialValue)
                self.widget.set(param, newValue)

            # Tuple animation
            for param, initialTuple in initialTupleParams.items():
                target = self.params[param]
                newValue = tuple(initialTuple[i] + step * (target[i] - initialTuple[i]) for i in range(len(initialTuple)))
                self.widget.set(param, newValue)

        # Ensure value is exactly correct at end
        for param, target in self.params.items():
            self.widget.set(param, target)

        self.started = False


class Translate(AnimationBase):
    def __init__(self, widget, timeout, x, y):
        super().__init__(widget, timeout, x=x, y=y)


class Resize(AnimationBase):
    def __init__(self, widget, timeout, width, height):
        super().__init__(widget, timeout, width=width, height=height)


class Recolour(AnimationBase):
    def __init__(self, widget, timeout, colour):
        super().__init__(widget, timeout, colour=colour)


if __name__ == '__main__':
    from pygame_widgets import Button
    def animate():
        resize.start()
        translate.start()


    pygame.init()
    win = pygame.display.set_mode((600, 600))

    button = Button(win, 100, 100, 300, 150, inactiveColour=(0, 200, 0), hoverColour=(0, 200, 0))

    resize = Resize(button, 3, 200, 200)
    translate = Recolour(button, 5, (0, 100, 100))
    button.setOnClick(animate)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        button.listen(events)
        button.draw()

        pygame.display.update()



from abc import abstractmethod, ABC


class WidgetBase(ABC):
    def __init__(self, win, x, y, width, height):
        """ Base for all widgets

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of button
        :type width: int
        :param height: Height of button
        :type height: int
        """
        self.win = win
        self._x = x
        self._y = y
        self._width = width
        self._height = height

        self._hidden = False

    @abstractmethod
    def listen(self, events):
        pass

    @abstractmethod
    def draw(self):
        pass

    def contains(self, x, y):
        return self._x < x < self._x + self._width and self._y < y < self._y + self._height

    def hide(self):
        self._hidden = True

    def show(self):
        self._hidden = False

    def moveX(self, x):
        self._x += x

    def moveY(self, y):
        self._y += y

    def get(self, attr):
        """Default setter for any attributes. Call super if overriding

        :param attr: Attribute to get
        :return: Value of the attribute
        """
        if attr == 'x':
            return self._x

        if attr == 'y':
            return self._y

        if attr == 'width':
            return self._width

        if attr == 'height':
            return self._height

    def getX(self):
        return self._x

    def getY(self):
        return self._y

    def getWidth(self):
        return self._width

    def getHeight(self):
        return self._height

    def set(self, attr, value):
        """Default setter for any attributes. Call super if overriding

        :param attr: Attribute to set
        :param value: Value to set
        """
        if attr == 'x':
            self._x = value

        if attr == 'y':
            self._y = value

        if attr == 'width':
            self._width = value

        if attr == 'height':
            self._height = value

    def setX(self, x):
        self._x = x

    def setY(self, y):
        self._y = y

    def setWidth(self, width):
        self._width = width

    def setHeight(self, height):
        self._height = height



import pygame
import time

from pygame_widgets.widget import WidgetBase


class TextBox(WidgetBase):
    # Times in ms
    REPEAT_DELAY = 400
    REPEAT_INTERVAL = 70
    CURSOR_INTERVAL = 400

    def __init__(self, win, x, y, width, height, **kwargs):
        """ A customisable textbox for Pygame

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of button
        :type width: int
        :param height: Height of button
        :type height: int
        :param kwargs: Optional parameters
        """
        super().__init__(win, x, y, width, height)

        self.selected = False
        self.showCursor = False
        self.cursorTime = 0
        self.cursorPosition = 0

        self.repeatTime = 0
        self.repeatKey = None
        self.firstRepeat = True
        self.keyDown = False

        self.text = []

        self.escape = False

        self.maxLengthReached = False

        # Border
        self.borderThickness = kwargs.get('borderThickness', 3)
        self.borderColour = kwargs.get('borderColour', (0, 0, 0))
        self.radius = kwargs.get('radius', 0)

        # Colour
        self.colour = kwargs.get('colour', (220, 220, 220))

        # Text
        self.placeholderText = kwargs.get('placeholderText', '')
        self.textColour = kwargs.get('textColour', (0, 0, 0))
        self.fontSize = kwargs.get('fontSize', 20)
        self.font = kwargs.get('font', pygame.font.SysFont('sans-serif', self.fontSize))

        self.textOffsetBottom = self.fontSize // 3
        self.textOffsetLeft = self.fontSize // 3
        self.textOffsetRight = self.fontSize // 2
        self.cursorOffsetTop = self._height // 6

        # Function
        self.onSubmit = kwargs.get('onSubmit', lambda *args: None)
        self.onSubmitParams = kwargs.get('onSubmitParams', ())

    def listen(self, events):
        """ Wait for inputs

        :param events: Use pygame.event.get()
        :type events: list of pygame.event.Event
        """
        if self.keyDown:
            self.updateRepeatKey()

        pressed = pygame.mouse.get_pressed()[0]
        x, y = pygame.mouse.get_pos()

        # Selection
        if pressed:
            if self.contains(x, y):
                self.selected = True
                self.showCursor = True
                self.cursorTime = time.time()

            else:
                self.selected = False
                self.showCursor = False
                self.cursorTime = time.time()

        # Keyboard Input
        if self.selected:
            for event in events:
                if event.type == pygame.KEYDOWN:
                    self.showCursor = True
                    self.keyDown = True
                    self.repeatKey = event
                    self.repeatTime = time.time()

                    if event.key == pygame.K_BACKSPACE:
                        if self.cursorPosition != 0:
                            self.maxLengthReached = False
                            self.text.pop(self.cursorPosition - 1)

                        self.cursorPosition = max(self.cursorPosition - 1, 0)

                    elif event.key == pygame.K_DELETE:
                        if not self.cursorPosition >= len(self.text):
                            self.maxLengthReached = False
                            self.text.pop(self.cursorPosition)

                    elif event.key == pygame.K_RETURN:
                        self.onSubmit(*self.onSubmitParams)

                    elif event.key == pygame.K_RIGHT:
                        self.cursorPosition = min(self.cursorPosition + 1, len(self.text))

                    elif event.key == pygame.K_LEFT:
                        self.cursorPosition = max(self.cursorPosition - 1, 0)

                    elif event.key == pygame.K_END:
                        self.cursorPosition = len(self.text)

                    elif event.key == pygame.K_ESCAPE:
                        if not self.escape:
                            self.selected = False
                            self.showCursor = False
                            self.escape = True
                            self.repeatKey = None
                            self.keyDown = None
                            self.firstRepeat = True

                    elif not self.maxLengthReached:
                        if len(event.unicode) > 0:
                            self.text.insert(self.cursorPosition, event.unicode)
                            self.cursorPosition += 1

                elif event.type == pygame.KEYUP:
                    self.repeatKey = None
                    self.keyDown = None
                    self.firstRepeat = True
                    self.escape = False

    def draw(self):
        """ Display to surface """
        if not self._hidden:
            if self.selected:
                self.updateCursor()

            borderRects = [
                (self._x + self.radius, self._y, self._width - self.radius * 2, self._height),
                (self._x, self._y + self.radius, self._width, self._height - self.radius * 2),
            ]

            borderCircles = [
                (self._x + self.radius, self._y + self.radius),
                (self._x + self.radius, self._y + self._height - self.radius),
                (self._x + self._width - self.radius, self._y + self.radius),
                (self._x + self._width - self.radius, self._y + self._height - self.radius)
            ]

            backgroundRects = [
                (
                    self._x + self.borderThickness + self.radius,
                    self._y + self.borderThickness,
                    self._width - 2 * (self.borderThickness + self.radius),
                    self._height - 2 * self.borderThickness
                ),
                (
                    self._x + self.borderThickness,
                    self._y + self.borderThickness + self.radius,
                    self._width - 2 * self.borderThickness,
                    self._height - 2 * (self.borderThickness + self.radius)
                )
            ]

            backgroundCircles = [
                (self._x + self.radius + self.borderThickness,
                 self._y + self.radius + self.borderThickness),
                (self._x + self.radius + self.borderThickness,
                 self._y + self._height - self.radius - self.borderThickness),
                (self._x + self._width - self.radius - self.borderThickness,
                 self._y + self.radius + self.borderThickness),
                (self._x + self._width - self.radius - self.borderThickness,
                 self._y + self._height - self.radius - self.borderThickness)
            ]

            for rect in borderRects:
                pygame.draw.rect(self.win, self.borderColour, rect)

            for circle in borderCircles:
                pygame.draw.circle(self.win, self.borderColour, circle, self.radius)

            for rect in backgroundRects:
                pygame.draw.rect(self.win, self.colour, rect)

            for circle in backgroundCircles:
                pygame.draw.circle(self.win, self.colour, circle, self.radius)

            x = [self._x + self.textOffsetLeft]
            for c in self.text:
                text = self.font.render(c, True, self.textColour)
                textRect = text.get_rect(bottomleft=(x[-1], self._y + self._height - self.textOffsetBottom))
                self.win.blit(text, textRect)
                x.append(x[-1] + text.get_width())

            if self.showCursor:
                try:
                    pygame.draw.line(
                        self.win, (0, 0, 0),
                        (x[self.cursorPosition], self._y + self.cursorOffsetTop),
                        (x[self.cursorPosition], self._y + self._height - self.cursorOffsetTop)
                    )
                except IndexError:
                    self.cursorPosition -= 1

            if x[self.cursorPosition] > self._x + self._width - self.textOffsetRight:
                self.maxLengthReached = True

    def updateCursor(self):
        now = time.time()

        if now - self.cursorTime >= self.CURSOR_INTERVAL / 1000:
            self.showCursor = not self.showCursor
            self.cursorTime = now

    def updateRepeatKey(self):
        now = time.time()

        if self.firstRepeat:
            if now - self.repeatTime >= self.REPEAT_DELAY / 1000:
                self.firstRepeat = False
                self.repeatTime = now
                pygame.event.post(
                    pygame.event.Event(
                        pygame.KEYDOWN,
                        {'key': self.repeatKey.key, 'unicode': self.repeatKey.unicode}
                    )
                )

        elif now - self.repeatTime >= self.REPEAT_INTERVAL / 1000:
            self.repeatTime = now
            pygame.event.post(
                pygame.event.Event(
                    pygame.KEYDOWN,
                    {'key': self.repeatKey.key, 'unicode': self.repeatKey.unicode}
                )
            )

    def setText(self, text):
        self.text = [c for c in str(text)]

    def getText(self):
        return ''.join(self.text)


if __name__ == '__main__':
    def output():
        print(textbox.getText())


    pygame.init()
    win = pygame.display.set_mode((1000, 600))

    textbox = TextBox(win, 100, 100, 800, 80, fontSize=50, borderColour=(255, 0, 0),
                      textColour=(0, 200, 0), onSubmit=output, radius=10,
                      borderThickness=5)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        textbox.listen(events)
        textbox.draw()

        pygame.display.update()



import pygame

from pygame_widgets.widget import WidgetBase


class Button(WidgetBase):
    def __init__(self, win, x, y, width, height, **kwargs):
        """ A customisable button for Pygame

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of button
        :type width: int
        :param height: Height of button
        :type height: int
        :param kwargs: Optional parameters
        """
        super().__init__(win, x, y, width, height)

        # Colour
        self.inactiveColour = kwargs.get('inactiveColour', (150, 150, 150))
        self.hoverColour = kwargs.get('hoverColour', (125, 125, 125))
        self.pressedColour = kwargs.get('pressedColour', (100, 100, 100))
        self.colour = self.inactiveColour
        self.shadowDistance = kwargs.get('shadowDistance', 0)
        self.shadowColour = kwargs.get('shadowColour', (210, 210, 180))

        # Function
        self.onClick = kwargs.get('onClick', lambda *args: None)
        self.onRelease = kwargs.get('onRelease', lambda *args: None)
        self.onClickParams = kwargs.get('onClickParams', ())
        self.onReleaseParams = kwargs.get('onReleaseParams', ())
        self.clicked = False

        # Text (Remove if using PyInstaller)
        self.textColour = kwargs.get('textColour', (0, 0, 0))
        self.fontSize = kwargs.get('fontSize', 20)
        self.string = kwargs.get('text', '')
        self.font = kwargs.get('font', pygame.font.SysFont('sans-serif', self.fontSize))
        self.text = self.font.render(self.string, True, self.textColour)
        self.textHAlign = kwargs.get('textHAlign', 'centre')
        self.textVAlign = kwargs.get('textVAlign', 'centre')
        self.margin = kwargs.get('margin', 20)

        self.textRect = self.text.get_rect()
        self.alignTextRect()

        # Image
        self.image = kwargs.get('image', None)
        self.imageHAlign = kwargs.get('imageHAlign', 'centre')
        self.imageVAlign = kwargs.get('imageVAlign', 'centre')

        if self.image:
            self.imageRect = self.image.get_rect()
            self.alignImageRect()

        # Shape
        self.radius = kwargs.get('radius', 0)

    def alignImageRect(self):
        self.imageRect.center = (self._x + self._width // 2, self._y + self._height // 2)

        if self.imageHAlign == 'left':
            self.imageRect.left = self._x + self.margin
        elif self.imageHAlign == 'right':
            self.imageRect.right = self._x + self._width - self.margin

        if self.imageVAlign == 'top':
            self.imageRect.top = self._y + self.margin
        elif self.imageVAlign == 'bottom':
            self.imageRect.bottom = self._y + self._height - self.margin

    def alignTextRect(self):
        self.textRect.center = (self._x + self._width // 2, self._y + self._height // 2)

        if self.textHAlign == 'left':
            self.textRect.left = self._x + self.margin
        elif self.textHAlign == 'right':
            self.textRect.right = self._x + self._width - self.margin

        if self.textVAlign == 'top':
            self.textRect.top = self._y + self.margin
        elif self.textVAlign == 'bottom':
            self.textRect.bottom = self._y + self._height - self.margin

    def listen(self, events):
        """ Wait for inputs

        :param events: Use pygame.event.get()
        :type events: list of pygame.event.Event
        """
        if not self._hidden:
            pressed = pygame.mouse.get_pressed()[0]
            x, y = pygame.mouse.get_pos()

            if self.contains(x, y):
                if pressed:
                    self.colour = self.pressedColour
                    if not self.clicked:
                        self.clicked = True
                        self.onClick(*self.onClickParams)

                elif self.clicked:
                    self.clicked = False
                    self.onRelease(*self.onReleaseParams)

                else:
                    self.colour = self.hoverColour

            elif not pressed:
                self.clicked = False
                self.colour = self.inactiveColour

    def draw(self):
        """ Display to surface """
        if not self._hidden:
            if pygame.version.vernum[0] < 2:
                rects = [
                    (self._x + self.radius, self._y, self._width - self.radius * 2, self._height),
                    (self._x, self._y + self.radius, self._width, self._height - self.radius * 2)
                ]

                circles = [
                    (self._x + self.radius, self._y + self.radius),
                    (self._x + self.radius, self._y + self._height - self.radius),
                    (self._x + self._width - self.radius, self._y + self.radius),
                    (self._x + self._width - self.radius, self._y + self._height - self.radius)
                ]

                for rect in rects:
                    x, y, width, height = rect
                    pygame.draw.rect(
                        self.win, self.shadowColour, (x + self.shadowDistance, y + self.shadowDistance, width, height)
                    )

                for circle in circles:
                    x, y = circle
                    pygame.draw.circle(
                        self.win, self.shadowColour, (x + self.shadowDistance, y + self.shadowDistance), self.radius
                    )

                for rect in rects:
                    pygame.draw.rect(self.win, self.colour, rect)

                for circle in circles:
                    pygame.draw.circle(self.win, self.colour, circle, self.radius)
            else:
                pygame.draw.rect(
                    self.win, self.shadowColour,
                    (self._x + self.shadowDistance, self._y + self.shadowDistance, self._width, self._height),
                    border_radius=self.radius
                )

                pygame.draw.rect(
                    self.win, self.colour, (self._x, self._y, self._width, self._height),
                    border_radius=self.radius
                )

            if self.image:
                self.win.blit(self.image, self.imageRect)

            self.win.blit(self.text, self.textRect)

    def setImage(self, image):
        self.image = image
        self.imageRect = self.image.get_rect()
        self.alignImageRect()

    def setOnClick(self, onClick, params=()):
        self.onClick = onClick
        self.onClickParams = params

    def setOnRelease(self, onRelease, params=()):
        self.onRelease = onRelease
        self.onReleaseParams = params

    def setInactiveColour(self, colour):
        self.inactiveColour = colour

    def setPressedColour(self, colour):
        self.pressedColour = colour

    def setHoverColour(self, colour):
        self.hoverColour = colour

    def get(self, attr):
        parent = super().get(attr)
        if parent is not None:
            return parent

        if attr == 'colour':
            return self.colour

    def set(self, attr, value):
        super().set(attr, value)

        if attr == 'colour':
            self.inactiveColour = value


class ButtonArray(WidgetBase):
    def __init__(self, win, x, y, width, height, shape, **kwargs):
        """ A collection of buttons

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of button
        :type width: int
        :param height: Height of button
        :type height: int
        :param shape: The 2d shape of the array (columns, rows)
        :type shape: tuple of int
        :param kwargs: Optional parameters
        """
        super().__init__(win, x, y, width, height)

        self.shape = shape
        self.numButtons = shape[0] * shape[1]

        # Array
        self.colour = kwargs.get('colour', (210, 210, 180))
        self.border = kwargs.get('border', 10)
        self.topBorder = kwargs.get('topBorder', self.border)
        self.bottomBorder = kwargs.get('bottomBorder', self.border)
        self.leftBorder = kwargs.get('leftBorder', self.border)
        self.rightBorder = kwargs.get('rightBorder', self.border)
        self.borderRadius = kwargs.get('borderRadius', 0)
        self.separationThickness = kwargs.get('separationThickness', self.border)

        self.buttonAttributes = {
            # Colour
            'inactiveColour': kwargs.get('inactiveColours', None),
            'hoverColour': kwargs.get('hoverColours', None),
            'pressedColour': kwargs.get('pressedColours', None),
            'shadowDistance': kwargs.get('shadowDistances', None),
            'shadowColour': kwargs.get('shadowColours', None),

            # Function
            'onClick': kwargs.get('onClicks', None),
            'onRelease': kwargs.get('onReleases', None),
            'onClickParams': kwargs.get('onClickParams', None),
            'onReleaseParams': kwargs.get('onReleaseParams', None),

            # Text
            'textColour': kwargs.get('textColours', None),
            'fontSize': kwargs.get('fontSizes', None),
            'text': kwargs.get('texts', None),
            'font': kwargs.get('fonts', None),
            'textHAlign': kwargs.get('textHAligns', None),
            'textVAlign': kwargs.get('textVAligns', None),
            'margin': kwargs.get('margins', None),

            # Image
            'image': kwargs.get('images', None),
            'imageHAlign': kwargs.get('imageHAligns', None),
            'imageVAlign': kwargs.get('imageVAligns', None),
            'imageRotation': kwargs.get('imageRotations', None),
            'imageFill': kwargs.get('imageFills', None),
            'imageZoom': kwargs.get('imageZooms', None),
            'radius': kwargs.get('radii', None)
        }

        self.buttons = []
        self.createButtons()

    def createButtons(self):
        across, down = self.shape
        width = (self._width - self.separationThickness * (across - 1) - self.leftBorder - self.rightBorder) // across
        height = (self._height - self.separationThickness * (down - 1) - self.topBorder - self.bottomBorder) // down

        count = 0
        for i in range(across):
            for j in range(down):
                x = self._x + i * (width + self.separationThickness) + self.leftBorder
                y = self._y + j * (height + self.separationThickness) + self.topBorder
                self.buttons.append(Button(self.win, x, y, width, height,
                                           **{k: v[count] for k, v in self.buttonAttributes.items() if v is not None})
                                    )
                count += 1

    def listen(self, events):
        """ Wait for inputs

        :param events: Use pygame.event.get()
        :type events: list of pygame.event.Event
        """
        if not self._hidden:
            for button in self.buttons:
                button.listen(events)

    def draw(self):
        """ Display to surface """
        if not self._hidden:
            rects = [
                (self._x + self.borderRadius, self._y, self._width - self.borderRadius * 2, self._height),
                (self._x, self._y + self.borderRadius, self._width, self._height - self.borderRadius * 2)
            ]

            circles = [
                (self._x + self.borderRadius, self._y + self.borderRadius),
                (self._x + self.borderRadius, self._y + self._height - self.borderRadius),
                (self._x + self._width - self.borderRadius, self._y + self.borderRadius),
                (self._x + self._width - self.borderRadius, self._y + self._height - self.borderRadius)
            ]

            for rect in rects:
                pygame.draw.rect(self.win, self.colour, rect)

            for circle in circles:
                pygame.draw.circle(self.win, self.colour, circle, self.borderRadius)

            for button in self.buttons:
                button.draw()

    def getButtons(self):
        return self.buttons


if __name__ == '__main__':
    pygame.init()
    win = pygame.display.set_mode((600, 600))

    button = Button(win, 100, 100, 300, 150, text='Hello', fontSize=50, margin=20,
                    inactiveColour=(255, 0, 0), pressedColour=(0, 255, 0), radius=20,
                    onClick=lambda: print('Click'), font=pygame.font.SysFont('calibri', 10),
                    textVAlign='bottom', imageHAlign='centre', imageVAlign='centre',
                    onRelease=lambda: print('Release'), shadowDistance=5)

    buttonArray = ButtonArray(win, 50, 50, 500, 500, (2, 2), border=100, texts=('1', '2', '3', '4'))

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        button.listen(events)
        button.draw()

        pygame.display.update()



import pygame
import math

from pygame_widgets.widget import WidgetBase


class Checkbox(WidgetBase):
    def __init__(self, win, x, y, width, height, items, **kwargs):
        """ A list of buttons that allows multiple selections

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of list
        :type width: int
        :param height: Height of list
        :type height: int
        :param items: Names of list items
        :type items: tuple of str
        :param kwargs: Optional parameters
        """
        super().__init__(win, x, y, width, height)

        self.items = items
        self.rows = len(items)
        self.rowHeight = self._height // self.rows
        self.selected = [False for _ in range(self.rows)]

        # Border
        self.borderThickness = kwargs.get('borderThickness', 3)
        self.borderColour = kwargs.get('borderColour', (0, 0, 0))
        self.radius = kwargs.get('radius', 0)

        # Checkbox
        self.boxSize = int(kwargs.get('boxSize', self._height / self.rows // 3))
        self.boxThickness = kwargs.get('boxThickness', 3)
        self.boxColour = kwargs.get('boxColour', (0, 0, 0))
        # TODO: selected image (tick) / colour

        # Colour
        self.colour = kwargs.get('colour', (255, 255, 255))

        # Alternating colours: overrides colour
        self.colour1 = kwargs.get('colour1', self.colour)
        self.colour2 = kwargs.get('colour2', self.colour)

        # Text
        self.textColour = kwargs.get('textColour', (0, 0, 0))
        self.fontSize = kwargs.get('fontSize', 20)
        self.font = kwargs.get('font', pygame.font.SysFont('sans-serif', self.fontSize))
        self.texts = [self.font.render(self.items[row], True, self.textColour) for row in range(self.rows)]
        self.textRects = self.createTextRects()

        self.clicked = False

        self.boxes = self.createBoxLocations()

    def createTextRects(self):
        textRects = []
        for row in range(self.rows):
            textRects.append(
                self.texts[row].get_rect(
                    center=(
                        self._x + self.boxSize * 2 + (self._width - self.boxSize * 2) // 2,
                        self._y + self.rowHeight * row + self.rowHeight // 2
                    )
                )
            )

        return textRects

    def createBoxLocations(self):
        boxes = []
        for row in range(self.rows):
            boxes.append(pygame.Rect(
                self._x + self.boxSize,
                self._y + self.rowHeight * row + self.boxSize,
                self.boxSize, self.boxSize
            ))
        return boxes

    def listen(self, events):
        """ Wait for inputs

        :param events: Use pygame.event.get()
        :type events: list of pygame.event.Event
        """
        if not self._hidden:
            pressed = pygame.mouse.get_pressed()[0]
            x, y = pygame.mouse.get_pos()

            if self.contains(x, y):
                if pressed:
                    if not self.clicked:
                        self.clicked = True
                        for row in range(self.rows):
                            if self.boxes[row].collidepoint(x, y):
                                self.selected[row] = not self.selected[row]

                elif self.clicked:
                    self.clicked = False

            elif not pressed:
                self.clicked = False

    def draw(self):
        """ Display to surface """
        if not self._hidden:
            for row in range(self.rows):
                colour = self.colour1 if not row % 2 else self.colour2
                if pygame.version.vernum[0] < 2:
                    pygame.draw.rect(
                        self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight)
                    )
                else:
                    if row == 0:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight),
                            border_top_left_radius=self.radius, border_top_right_radius=self.radius
                        )

                    elif row == self.rows - 1:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight),
                            border_bottom_left_radius=self.radius, border_bottom_right_radius=self.radius
                        )

                    else:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight)
                        )

                width = 0 if self.selected[row] else self.boxThickness
                pygame.draw.rect(
                    self.win, self.boxColour,
                    self.boxes[row],
                    width
                )

                self.win.blit(self.texts[row], self.textRects[row])

    def getSelected(self):
        return [self.items[row] for row in range(self.rows) if self.selected[row]]


class Radio(WidgetBase):
    def __init__(self, win, x, y, width, height, items, **kwargs):
        """ A list of buttons that allows a single selections

        :param win: Surface on which to draw
        :type win: pygame.Surface
        :param x: X-coordinate of top left
        :type x: int
        :param y: Y-coordinate of top left
        :type y: int
        :param width: Width of list
        :type width: int
        :param height: Height of list
        :type height: int
        :param items: Names of list items
        :type items: tuple of str
        :param kwargs: Optional parameters
        """
        super().__init__(win, x, y, width, height)

        self.items = items
        self.rows = len(items)
        self.rowHeight = self._height // self.rows
        self.selected = kwargs.get('default', 0)

        # Border
        self.borderThickness = kwargs.get('borderThickness', 3)
        self.borderColour = kwargs.get('borderColour', (0, 0, 0))
        self.radius = kwargs.get('radius', 0)

        # Radio
        self.circleRadius = int(kwargs.get('circleRadius', self._height / self.rows // 6))
        self.circleThickness = kwargs.get('circleThickness', 3)
        self.circleColour = kwargs.get('circleColour', (0, 0, 0))

        # Colour
        self.colour = kwargs.get('colour', (255, 255, 255))

        # Alternating colours: overrides colour
        self.colour1 = kwargs.get('colour1', self.colour)
        self.colour2 = kwargs.get('colour2', self.colour)

        # Text
        self.textColour = kwargs.get('textColour', (0, 0, 0))
        self.fontSize = kwargs.get('fontSize', 20)
        self.font = kwargs.get('font', pygame.font.SysFont('sans-serif', self.fontSize))
        self.texts = [self.font.render(self.items[row], True, self.textColour) for row in range(self.rows)]
        self.textRects = self.createTextRects()

        self.clicked = False

        self.circles = self.createCircleLocations()

    def createTextRects(self):
        textRects = []
        for row in range(self.rows):
            textRects.append(
                self.texts[row].get_rect(
                    center=(
                        self._x + self.circleRadius * 6 + (self._width - self.circleRadius * 6) // 2,
                        self._y + self.rowHeight * row + self.rowHeight // 2
                    )
                )
            )

        return textRects

    def createCircleLocations(self):
        circles = []
        for row in range(self.rows):
            circles.append((
                self._x + self.circleRadius * 3,
                self._y + self.rowHeight * row + self.rowHeight // 2,
            ))
        return circles

    def listen(self, events):
        """ Wait for inputs

        :param events: Use pygame.event.get()
        :type events: list of pygame.event.Event
        """
        if not self._hidden:
            pressed = pygame.mouse.get_pressed()[0]
            x, y = pygame.mouse.get_pos()

            if self.contains(x, y):
                if pressed:
                    if not self.clicked:
                        self.clicked = True
                        for row in range(self.rows):
                            if math.sqrt((self.circles[row][0] - x) ** 2 +
                                         (self.circles[row][1] - y) ** 2) <= self.circleRadius:
                                self.selected = row

                elif self.clicked:
                    self.clicked = False

            elif not pressed:
                self.clicked = False

    def draw(self):
        """ Display to surface """
        if not self._hidden:
            for row in range(self.rows):
                colour = self.colour1 if not row % 2 else self.colour2
                if pygame.version.vernum[0] < 2:
                    pygame.draw.rect(
                        self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight)
                    )

                else:
                    if row == 0:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight),
                            border_top_left_radius=self.radius, border_top_right_radius=self.radius
                        )

                    elif row == self.rows - 1:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight),
                            border_bottom_left_radius=self.radius, border_bottom_right_radius=self.radius
                        )

                    else:
                        pygame.draw.rect(
                            self.win, colour, (self._x, self._y + self.rowHeight * row, self._width, self.rowHeight)
                        )

                width = 0 if row == self.selected else self.circleThickness
                pygame.draw.circle(
                    self.win, self.circleColour,
                    self.circles[row], self.circleRadius,
                    width
                )

                self.win.blit(self.texts[row], self.textRects[row])


if __name__ == '__main__':
    pygame.init()
    win = pygame.display.set_mode((1000, 800))

    checkbox = Checkbox(win, 100, 100, 400, 300, ('Apples', 'Bananas', 'Pears'), colour1=(0, 180, 0),
                        colour2=(0, 50, 200), fontSize=30, radius=10)
    radio = Radio(win, 550, 400, 400, 300, ('Apples', 'Bananas', 'Pears'), colour1=(0, 180, 0),
                  colour2=(0, 50, 200), fontSize=30, radius=10)

    run = True
    while run:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                run = False
                quit()

        win.fill((255, 255, 255))

        checkbox.listen(events)
        checkbox.draw()

        radio.listen(events)
        radio.draw()

        pygame.display.update()
